<div class="ui secondary segment" style="max-height: 300px; overflow-y: auto;">
    <div class="ui relaxed divided list" data-widget="log.log_list" data-spec='{{ spec|tojson }}'
         data-interval="{{ interval }}">
    </div>
</div>
